#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>
#include <pthread.h>
#include <libgen.h>	/* basename */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/stat.h>

#include <xmlconfig.h>

#include <pv_types.h>
#include <pv_bs.h>
#include <pv_bs_8bit.h>
#include <pv_lock.h>
#include <jv_server.h>

static const char *prompt = "pv";


#ifdef _SEM_SEMUN_UNDEFINED
union semun {
	int val;
	struct semid_ds *buf;
	unsigned short int *array;
	struct seminfo *__buf;
};
#endif

#define D1(x) if (pv_verbose) x
extern int pv_verbose;


static void show_help(char *name)
{
	fprintf(stderr, "\nusage:   "
		"%s [-v] -d <file> -s <file> [-j <port>]\n\n"
		"  -v, --verbose             Verbose mode of operation\n"
		"  -d, --description <file>  Description XML File: specifies layout of the \n"
		"                            process variable store file\n"
		"  -s, --store <file>        Store file: this file holds the process variables\n"
		"  -p, --pv8bit              use 8 bit backing store\n"
		"  -j, --jvisu <port>        Start JVisu Socket Server\n"
		"\n", name);
}


pv_engine_t *pv_init(int argc, char *argv[])
{
	int pv_store_size; 
	pv_t *tmp_pv = NULL; 
	xmlconfig_t *xmlcfg;
	int i,ret;
	pv_verbose = 0;

	/* xml parser */
	int no_pvs;
	char buf[1024];

	/* option parser */
	struct option long_options[] = {
		{ "help",        no_argument,       0, 'h' },
		{ "verbose",     no_argument,       0, 'v' },
		{ "description", required_argument, 0, 'd'}, 
		{ "store",       required_argument, 0, 's'},
		{ "pv8bit",      no_argument,       0, 'p'},
		{ 0,             0,                 0,  0 },
	};
	int longindex;
	int opt;

	/* create and initialize new context */
	struct pv_engine *pv_engine = calloc(1,sizeof(struct pv_engine));
	if (pv_engine == NULL) {
		fprintf(stderr, "%s: %s: error: out of memory\n", prompt, __FUNCTION__);
		return NULL;
	}
	pv_engine->jv_varlist = NULL;
	pv_engine->store.init = pv_bs_init;
	pv_engine->pv_shmid = -1;

	/* reset getopt */
	optind = 0;

	/* parse command line switches */
	while ((opt = getopt_long(argc, argv, "hvd:s:j:p", long_options, &longindex)) != -1) {
		switch (opt) {
		case 'h':
			show_help(basename(argv[0]));
			free(pv_engine);
			return NULL;
		case 'v':
			pv_verbose++;
			break;	
		case 'd':
			pv_engine->xml_filename=optarg;
			break;
		case 's':
			pv_engine->pv_filename=optarg;
			break;
		case 'j':
			pv_engine->jv_port=strtoul(optarg,NULL,0);
			break;
		case 'p':
			pv_engine->store.init = pv_bs_8bit_init;
			break;
                default:
                        show_help(argv[0]);
			free(pv_engine);
			return NULL;
                }
        }

	/* parse command line options */
	if (((argc - optind) != 0) ||
	    (pv_engine->xml_filename == NULL)) {
		show_help(argv[0]);
		free(pv_engine);
		return NULL;
	}

	if ( !pv_engine->pv_filename ) {
		fprintf( stderr, "You must specify a store file using the -s option\n");
		return NULL;
	}

	VFPRINTF(1, stdout, "\n"
		         "description file...: %s\n"
			 "store file.........: %s\n"
			 "\n" , pv_engine->xml_filename, pv_engine->pv_filename);
	
	/* analyse xml description file */
	xmlcfg = xmlconfig_init(pv_engine->xml_filename);
	if (!xmlcfg) {
		fprintf(stderr, "could not open xml file \"%s\": ", pv_engine->xml_filename);
		perror("");
		free(pv_engine);
		return NULL;
	}

	no_pvs = xmlconfig_get_string(xmlcfg, "//pv", &buf[0], 0, 0);
	/* FIXME: exit if no pvs */
	pv_engine->pv_list = list_init();
	for (i=0; i < no_pvs; i++) {
		pv_t pv;
		char xpath[80];
		int err;
		unsigned long tmp;

		/* id */
		snprintf(&xpath[0], 80, "//pv[%i]/@id", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		pv.name = malloc(strlen(buf)+1);
		if (!pv.name)
			fprintf(stderr, "error: could not get memory\n");
		strncpy(pv.name, buf, strlen(buf)+1);
		
		/* type */
		snprintf(&xpath[0], 80, "//pv[%i]/type", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		if (strncmp(buf,"uint8_t",sizeof(buf)) == 0)
			pv.type = pv_type_uint8;
		if (strncmp(buf,"uint16_t",sizeof(buf)) == 0)
			pv.type = pv_type_uint16;
		if (strncmp(buf,"uint32_t",sizeof(buf)) == 0)
			pv.type = pv_type_uint32;
		if (strncmp(buf,"string",sizeof(buf)) == 0)
			pv.type = pv_type_string;

		/* offset */
		snprintf(&xpath[0], 80, "//pv[%i]/offset", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		/* FIXME: exit if no offset */
		/* grmpf: please _NO_ exit()-calls in libs! (mkl) */
		pv.offset = strtoul(buf, NULL, 0);

		/* desc */
		snprintf(&xpath[0], 80, "//pv[%i]/desc", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		pv.desc = malloc(strlen(buf)+1);
		if (!pv.desc)
			fprintf(stderr, "error: could not get memory\n");
		strncpy(pv.desc, buf, strlen(buf)+1);

		/* size */
		snprintf(&xpath[0], 80, "//pv[%i]/size", i+1);
		err = xmlconfig_get_ulong(xmlcfg, &xpath[0], (unsigned long *)(&pv.size), 1);
		if (err == 0) {		/* no match, thus no size :) */
			switch (pv.type) {
			case pv_type_uint8:
				pv.size = 1;
				break;
			case pv_type_uint16:
				pv.size = 2;
				break;
			case pv_type_uint32:
				pv.size = 4;
				break;
			case pv_type_string:
				exit(EXIT_FAILURE);	/* <---- this seems to be usual RSC practice :) (mkl) */
				break;
			case pv_type_undef:
				exit(EXIT_FAILURE);	/* <---- this seems to be usual RSC practice :) (mkl) */
				break;
			}
		}
		
		/* default */
		snprintf(&xpath[0], 80, "//pv[%i]/default", i+1);
		if (pv.type == pv_type_uint8
		    || pv.type == pv_type_uint16
		    || pv.type == pv_type_uint32) {
			/* FIXME: HACK, fix xmlconfig fist, then here (mkl) */
			err = xmlconfig_get_ulong(xmlcfg, &xpath[0], &tmp, 1);
			switch (pv.type) {
			case pv_type_uint8:
				*(uint8_t *) (&pv.dflt) = tmp;
				break;
			case pv_type_uint16:
				*(uint16_t *)(&pv.dflt) = tmp;
				break;
			case pv_type_uint32:
				*(uint32_t *)(&pv.dflt) = tmp;
				break;
			case pv_type_string:
			case pv_type_undef:
				break;
			}
			if (err == 0)
				memset(&pv.dflt, 0, sizeof(pv.dflt));
		} else if (pv.type == pv_type_string) {
			pv.dflt = calloc(1, pv.size);
			xmlconfig_get_string(xmlcfg, &xpath[0], (char *)pv.dflt, pv.size, 1);
		}
			
		/* scale */
		snprintf(&xpath[0], 80, "//pv[%i]/scale", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		/* FIXME: not implemented yet */

		/* unit */
		snprintf(&xpath[0], 80, "//pv[%i]/unit", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		/* FIXME: not implemented yet */

		/* max */
		snprintf(&xpath[0], 80, "//pv[%i]/max", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		/* FIXME: not implemented yet */

		/* min */
		snprintf(&xpath[0], 80, "//pv[%i]/min", i+1);
		xmlconfig_get_string(xmlcfg, &xpath[0], &buf[0], 1024, 1);
		/* FIXME: not implemented yet */

		pv_engine->pv_list = list_add_tail(pv_engine->pv_list, (void*)&pv, sizeof(pv_t));
		
	}

	/* tell us which process variables have been found */
	if ((is_empty(pv_engine->pv_list)) && pv_verbose) {
		
		pv_t *pv;
		int finished = 0; 

		printf("process variables:\n");
		printf("------------------\n\n");
		
		for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
			pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
			
			if (is_last(pv_engine->pv_list))
				finished=1;

			printf("n:%s o=%li\n",pv->name, pv->offset);
		}
	}

	/* find out size of pv area */
	pv_store_size=0;
	if (is_last(pv_engine->pv_list) != -1) {
	
		int finished = 0; 

		for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
			
			pv_t *pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
			
			if (is_last(pv_engine->pv_list))
				finished=1;
			
			if (pv->offset > pv_store_size) {
				pv_store_size = pv->offset;
				tmp_pv = pv; 
			}
		}
	}

	pv_store_size += tmp_pv->size;

	VFPRINTF(1, stderr, "pv store size = %i\n", pv_store_size); 

	/* locking */
	/* FIXME: use absolute paths */
	pv_engine->pv_sem_key = ftok(pv_engine->pv_filename, 42); 
	pv_engine->pv_sem = semget(pv_engine->pv_sem_key, 1, 0666 | IPC_CREAT | IPC_EXCL);
	if( pv_engine->pv_sem < 0) {
		pv_engine->pv_sem = semget(pv_engine->pv_sem_key, 1, 0666 | IPC_CREAT);
		if (pv_engine->pv_sem < 0) {
			fprintf(stderr, "error: could not create semaphore\n");
			perror("error: ");
			exit(EXIT_FAILURE);
		}
	} else {
		union semun sem_val; 
		sem_val.val = 1; 
		if (semctl(pv_engine->pv_sem, 0, SETVAL, sem_val) < 0) {
			fprintf(stderr, "error: could not initialize semaphore\n");
			perror("error: ");
			exit(EXIT_FAILURE);
		}
	}

	_pv_lock(pv_engine);

	VFPRINTF(1, stdout, "\n%s: %s: sem_id=%i\n", prompt, __FUNCTION__, pv_engine->pv_sem);

	pv_engine->store.size = pv_store_size;
	ret = pv_engine->store.init(&pv_engine->store, pv_engine->pv_filename);
	if(ret < 0) {
		fprintf(stderr, "Could not initialize backing store\n");
		exit(EXIT_FAILURE);
	}

	if (!pv_engine->store.map) {
		int pv_init_store = 0;
		VFPRINTF(1, stderr,"opening shared memory segment\n");

		pv_engine->pv_shmid = shmget(pv_engine->pv_sem_key, pv_store_size, IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR );
		if(pv_engine->pv_shmid < 0) {
			VFPRINTF(1, stderr,"shared memory segment already exists. attaching...\n");
			pv_engine->pv_shmid = shmget(pv_engine->pv_sem_key, pv_store_size, IPC_CREAT | S_IRUSR | S_IWUSR );
			if(pv_engine->pv_shmid < 0) {
				perror("shmget");
				exit(EXIT_FAILURE);			
			}
		} else
			pv_init_store = 1;

		pv_engine->store.map = shmat(pv_engine->pv_shmid, NULL, 0);
		if(!pv_engine->store.map) {
			perror("shmat");
			exit(1);
		}

		if(pv_init_store) {
			VFPRINTF(1, stderr, "initialising store\n");
			pv_engine->store.read(&pv_engine->store, 0, pv_store_size);
		}
	}

	pv_engine->pv_map = pv_engine->store.map;
	pv_engine->pv_store_size = pv_store_size;

	if (pv_engine->jv_port != 0) {
	
		/* create jvisu socket thread; this may not work if another process has already started one */
		VFPRINTF(1, stdout, "%s: %s: starting JVisu socket server on port %i\n", prompt, __FUNCTION__, pv_engine->jv_port);
		if (pthread_create(&pv_engine->jvsocket_thread, 0, pv_jvsocket_task, (void*)pv_engine) != 0) {
			fprintf(stderr, "warning: could not create socket thread\n");
		}

	} else 
		VFPRINTF(1, stdout, "%s: %s: -j not given, so don't start JVisu socket server\n", prompt, __FUNCTION__);	
	
	VFPRINTF(1, stdout, "\n");
	
 	_pv_unlock(pv_engine);
	return (pv_engine_t *)pv_engine;
}


void pv_finish (pv_engine_t *pv_engine_anon)
{
	struct pv_engine *pv_engine = (struct pv_engine *)pv_engine_anon;

	/* shutdown jvisu socket server */
	D1(printf("%s: %-20s: stopping server\n", prompt, __FUNCTION__));
	pv_stop_jvserver(pv_engine, 1);
	
	if (pv_engine->jvsocket_thread) {
		D1(printf("%s: %-20s: waiting for server thread\n", prompt, __FUNCTION__));
		pthread_join(pv_engine->jvsocket_thread, 0);
		D1(printf("%s: %-20s: joined\n", prompt, __FUNCTION__));
	}

	/* FIXME: do we need to remove the SysV semaphore? */
	/* SHA: Yes, but make sure we are the last process using it */

	if(pv_engine->pv_shmid >= 0) {
		VFPRINTF(1, stderr, "detaching shared memory\n");
		shmdt(pv_engine->pv_map);
		shmctl(pv_engine->pv_shmid, IPC_RMID, NULL);
	}

	/* remove process variables */
	VFPRINTF(1, stdout, "freeing pv list:");

	if (is_last(pv_engine->pv_list) != -1) {

		int finished = 0; 
	
		for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
			
			pv_t *pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
			
			if (is_last(pv_engine->pv_list))
				finished=1;

			if (pv == NULL) {
				fprintf(stderr, "error: trying to free nonexistent list element\n");
				exit(EXIT_FAILURE);
			}
			free(pv->name);
			free(pv->desc);
			if (pv->type == pv_type_string)
				free(pv->dflt);

			VFPRINTF(1, stdout, ".");
			
			if (is_last(pv_engine->pv_list))
				goto end_finish;

			pv_engine->pv_list = list_del_element(pv_engine->pv_list);
		}
	}
	
end_finish:	
	VFPRINTF(1, stdout, "\n");

	return;
}
